# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)

# package metadata
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = '0.1a8'
__date__ = '2014-08-12 16:53:39 +0000'

# package version
version = '0.1a8'
major = 0
minor = 1
micro = None
debug = True
release = False

# repository version information
git_hash = 'e9e172c49e876d9e111b403059b328e9529c1c89'
git_tag = 'v0.1a8'
git_author = "Duncan Macleod <duncan.macleod@ligo.org>"
git_committer = "Duncan Macleod <duncan.macleod@ligo.org>"
git_is_dirty = False